from persistent.dict import PersistentDict
from persistent.list import PersistentList
from zope.interface import implements
from zope.component import adapts
from Products.ATContentTypes.interface.image import IImageContent
from Products.ImageEditor.interfaces import IImageEditorAdapter
from PIL import Image
from cStringIO import StringIO
from Acquisition import aq_inner


class ImageEditorAdapter(object):
    """context.getField('image').get(context) -> Field Image
    Field Image.data -> OFS Image
    OFSImage.data    -> str
    
    stack = [str, ...]
    """
    implements(IImageEditorAdapter)
    adapts(IImageContent)
    
    def __init__(self, context):
        self.context = aq_inner(context)

    @property
    def storage(self):
        session = self.context.REQUEST.SESSION      # evil! :)
        if not session.has_key('imageeditor'):
            session['imageeditor'] = PersistentDict()
        editordata = session['imageeditor']
        return editordata.setdefault(self.context.UID(), PersistentDict())

    @property
    def field(self):
        storage = self.storage
        fieldname = storage.get('fieldname')
        if fieldname:
            return self.context.getField(fieldname)
        else:
            return self.context.getField('image') or \
                self.context.getPrimaryField()

    def set_field(self, name):
        storage = self.storage
        storage['fieldname'] = name     # set name before clear

    @property
    def field_data(self):
        storage = self.storage
        return storage.setdefault(storage.get('fieldname'), PersistentDict())

    def get_image_data(self):
        return str(self.field.get(self.context).data)

    # UNDO REDO STUFF
    def get_pos(self):
        return self.field_data.setdefault('position', 0)
    def set_pos(self, value):
        self.field_data['position'] = value
    pos = property(get_pos, set_pos)

    @property
    def stack(self):
        stack = self.field_data.setdefault('unredostack', PersistentList())
        if not stack:
            stack.append(self.get_image_data())
        return stack

    def undo(self):
        if self.can_undo():
            self.pos = self.pos - 1

    def can_undo(self):
        return self.pos > 0

    def redo(self):
        if self.can_redo():
            self.pos = self.pos + 1

    def can_redo(self):
        return self.pos + 1 < len(self.stack)

    def clear_edits(self, bottom=None):
        if bottom is None:
            bottom = self.get_image_data()
        stack = self.stack
        stack[:] = [bottom]
        self.pos = 0

    def do(self, value):
        if self.can_redo():
            for item in self.stack[(self.pos+1):len(self.stack)]:
                self.stack.remove(item)

        self.stack.append(value)
        self.pos = self.pos + 1

    def save_edit(self):
        image_data = self.get_current_image_data()
        
        field = self.field
        mimetype = field.getContentType(self.context)
        filename = field.getFilename(self.context)
        
        # because AT tries to get mimetype and filename from a file like
        # object by attribute access I'm passing a string along
        field.set(
            self.context,
            image_data, 
            mimetype=mimetype,
            filename=filename, 
            refresh_exif=False
        )
        
        #should I clear???  Maybe you should still be able to undo...              
        self.clear_edits(image_data)

    def get_current_image(self):
        return Image.open(StringIO(self.get_current_image_data()))

    def get_current_image_data(self):
        return self.stack[self.pos]

    def get_current_image_info(self):
        data = self.get_current_image_data()
        bsize = len(data)
        bsize = bsize/1024

        size_descriptor = 'kb'
        if bsize > 1024:
            bsize = bsize/1024.0
            size_descriptor = 'mb'

        size = self.get_current_image().size

        width = size[0]
        height = size[1]

        return {
            'size': bsize,
            'width': width,
            'height': height,
            'sizeformatted': "Size: %.2f%s" % (bsize, size_descriptor)
        }


    def set_image(self, image, format="JPEG", quality=None):
        image_data = StringIO()
        
        if quality:
            image.save(image_data, format, quality=quality)
        else:
            image.save(image_data, format)
        
        self.do(image_data.getvalue())





from zope.app.form.browser.widget import SimpleInputWidget
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile

class SliderWidget(SimpleInputWidget):
    
    template = ViewPageTemplateFile('sliderwidget.pt')

    def id(self):
        return self.name.replace('.', '-')

    def action_id(self):
        return self._prefix.replace('.', '')

    def js(self):
        return """
(function($){
$(document).ready(function(){

    $("#%(id)s-slider").slider({
        min:%(min)i,
        max:%(max)i,
        value: %(default)i,
        change: function(e, ui){
            $("#%(id)s-value").attr('value', ui.value);
            $("#%(id)s-slider p").html(String(ui.value));
        }
    });
    
    on('action_button_clicked').accomplish(function(btn){
        if($(btn).attr('name') == '%(action_id)s'){
            $('#%(id)s-slider').slider('moveTo', %(default)i);
        }
    });
});
})(jQuery);

        """ % {
            'id' : self.id(),
            'min' : self.context.min,
            'max' : self.context.max,
            'default' : self.context.default,
            'action_id' : self.action_id()
        }

    def __call__(self):
        return self.template(self)



import unittest
from Products.ImageEditor.tests import base
from Products.ImageEditor.actions.actions import *

class TestActions(base.ImageEditorTestCase):
    """Test Actions"""
    def afterSetUp(self):
        super(TestActions, self).afterSetUp()
        self.image = self.getImageContentType()
        
    def beforeTearDown(self):
        parent = self.image.getParentNode()
        parent.manage_delObjects([self.image.getId()])

    def test_crop(self):
        action = CropAction(self.image)
        x1, y1, x2, y2 = 0, 0, 10, 10
        action(x1, y1, x2, y2)

    def test_rotateLeft(self):
        RotateLeftAction(self.image)()
    
    def test_blur(self):
        action = BlurAction(self.image)
        amount = 2 # min=0,max=8,
        action(amount)

    def test_save(self):
        SaveImageEditAction(self.image)()

    def test_cancel(self):
        CancelImageEditAction(self.image)()

    def test_redo(self):
        RedoAction(self.image)()

    def test_undo(self):
        UndoAction(self.image)()

    def test_rotateRight(self):
        RotateRightAction(self.image)()

    def test_flipVertical(self):
        FlipOnVerticalAxisAction(self.image)()

    def test_compress(self):
        action = CompressAction(self.image)
        amount = 60.0
        action(amount)

    def test_contrast(self):
        action = ContrastAction(self.image)
        amount = 50.0
        action(amount)

    def test_bright(self):
        action = BrightnessAction(self.image)
        amount = 50.0
        action(amount)
        
    def test_sharpen(self):
        action = SharpenAction(self.image)
        amount = 1.5
        action(amount)
        
    def test_flipHorizontal(self):
        FlipOnHorizontalAxisAction(self.image)()
        
    def test_resize(self):
        action = ResizeAction(self.image)
        width, height = 150, 150
        action(width, height)
        
    def test_dropShadow(self):
        action = DropShadowAction(self.image)
        offset_x = 5
        offset_y = 5
        background_color = u"ffffff"
        shadow_color = u"444444"
        border = 8
        iterations = 3
        action(offset_x, offset_y, background_color, shadow_color, border, iterations)
        
    def test_save_as_image(self):
        parent = self.image.getParentNode()
        action = SaveAsImageEditAction(self.image)
        title = "foobar"
        _type = "Image"
        res = action(_type, title)
        
        self.failUnless(title in parent.objectIds())
        self.failUnless(res['new_type_location'] == "%s/%s/edit" % (parent.absolute_url(), title))
        self.failUnless(parent['foobar'].portal_type == "Image")
    
    def test_save_as_news_item(self):
        parent = self.image.getParentNode()
        action = SaveAsImageEditAction(self.image)
        title = "foobar"
        _type = "News Item"
        res = action(_type, title)

        self.failUnless(title in parent.objectIds())
        self.failUnless(res['new_type_location'] == "%s/%s/edit" % (parent.absolute_url(), title))
        self.failUnless(parent['foobar'].portal_type == "News Item")

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestActions))
    return suite



from Products.ImageEditor.tests.base import ImageEditorTestCase
from Products.ImageEditor.actions import *

class TestAdapter(ImageEditorTestCase):
    """
    Test the edit actions on images
    """
        
    def test_undo(self):
        editor = self.getImageEditorAdapter()
        
        original_image = editor.get_current_image_data()
        RotateRightAction(editor.context)()
        self.failUnless(original_image != editor.get_current_image_data())
        
        editor.undo()
        self.failUnless(original_image == editor.get_current_image_data())
        
    def test_redo(self):
        editor = self.getImageEditorAdapter()
        
        original_image = editor.get_current_image_data()

        RotateRightAction(editor.context)()
        edited_image = editor.get_current_image_data()
        self.failUnless(original_image != edited_image)
        
        editor.undo()
        self.failUnless(original_image == editor.get_current_image_data())
        
        editor.redo()
        self.failUnless(edited_image == editor.get_current_image_data())
        
    def test_clear_edits(self):
        editor = self.getImageEditorAdapter()
        
        original_image = editor.get_current_image_data()
        
        RotateRightAction(editor.context)()
        RotateRightAction(editor.context)()
        RotateRightAction(editor.context)()
        
        editor.clear_edits()
        self.failUnless(original_image == editor.get_current_image_data())
        
    def test_save_edits(self):
        editor = self.getImageEditorAdapter()
        
        RotateRightAction(editor.context)()
        RotateRightAction(editor.context)()
        
        current = editor.get_current_image_data()
        editor.save_edit()
        
        self.failUnless(len(editor.stack) == 1)
        self.failUnless(current == editor.get_current_image_data())
        
    def test_get_current_image_info(self):
        editor = self.getImageEditorAdapter()
        info = editor.get_current_image_info()

        self.failUnless(info['sizeformatted'] == "Size: 25.00kb")
        self.failUnless(info['width'] == 461)
        self.failUnless(info['height'] == 614)
        self.failUnless(info['size'] == 25)
        
def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestAdapter))
    
    return suite


from Products.ImageEditor.interfaces.actions import IImageEditorAction
from Products.ImageEditor.interfaces.imageeditor import IImageEditorAdapter
from zope.interface import implements
from zope.component import adapts

class BaseImageEditorAction:
    adapts(IImageEditorAdapter)
    implements(IImageEditorAction)
    
    name = ""
    description = ""
    skip_apply = False
    options = None
    icon = None
    
    def __init__(self, image):
        self.editor = IImageEditorAdapter(image)
        
    def action_parameters(self):
        return False
    
    def on_setup(self):
        return False
    
    def __call__(self):
        raise Exception("Not Implemented")


from Products.ImageEditor.tests.base import ImageEditorTestCase
from Products.CMFCore.utils import getToolByName


class TestSetup(ImageEditorTestCase):
    """Test if install of the product is well done"""
    
    def test_actions(self):
        actionTool = self.portal.portal_actions
        actionInfo = actionTool.getActionInfo(['object/image_editor'])
        self.failUnless(actionInfo['url'] == "/@@imageeditor")

    def test_js_added(self):
        pjavascripts = getToolByName(self.portal, 'portal_javascripts')
        self.failUnless('++resource++jquery.imgareaselect-0.8.min.js' not in [js.getId() for js in pjavascripts.getResources()])

    def test_should_install_collective_js_jqueryui(self):
        qi = getToolByName(self.portal, 'portal_quickinstaller')
        self.failUnless(qi.isProductInstalled('collective.js.jqueryui'))

    def test_should_remove_action_object_on_uninstall_profile(self):
        self.uninstall()
        actionTool = self.portal.portal_actions
        object_actions = actionTool.object
        self.assertEquals('image_editor' in object_actions.objectIds(), False)
        
    def test_should_remove_javascript_on_uninstall_profile(self):
        self.uninstall()
        pjavascripts = getToolByName(self.portal, 'portal_javascripts')
        jsresources = [js.getId() for js in pjavascripts.getResources()]
        self.failUnless('++resource++jquery.imgareaselect-0.8.min.js' not in jsresources)

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestSetup))
    return suite



from base import BaseImageEditorAction
from zope.interface import implements
from Products.ImageEditor.interfaces.actions import IImageEditorAction
from options import INoOptions, IBlurOptions, ICompressOptions, IContrastOptions, \
    IBrightnessOptions, ISharpenOptions, IDropShadowOptions, ISepiaOptions, ISaveAsOptions
from zope.formlib import form
from PIL import Image, ImageFilter, ImageEnhance, ImageOps
from widgets import SliderWidget
from Products.CMFCore.utils import getToolByName
from zope.component import queryUtility
from plone.i18n.normalizer.interfaces import IURLNormalizer

class CropAction(BaseImageEditorAction):
    implements(IImageEditorAction)

    options = form.FormFields(INoOptions)
    name = u"Crop"
    description = u"Crop the image."
    icon = u"++resource++imageeditor/icons/stock-selection-intersect-16.png"
    
    def on_setup(self):
        return """
function add_crop(){
    var w = $('#source-image').width(), h = $('#source-image').height();
    start_selection = {
        x1: w/4,
        y1: h/4,
        x2: w-w/4,
        y2: h-h/4
    };
    window.crop_selection = start_selection;
    
    function create_cropper(ratio){
        var options = {
            onSelectChange: function(image, selection){
                window.crop_selection = selection;
            },
            selectionColor: 'blue',
            enable: true,
            border: 2,
            handles: true,
            show: true,
            parent: '#image-container',
            x1 : start_selection.x1,
            y1 : start_selection.y1,
            x2 : start_selection.x2,
            y2 : start_selection.y2
        }
        if(ratio != null){
            options.aspectRatio = ratio;
        }
        $('#source-image').imgAreaSelect(options);
    }
    create_cropper(null);
    
    $("#aspect-ratio-selector").dialog({
        position : 'right',
        title : "Aspect Ratio",
        autoOpen:true,
        resizable:false,
        modal: false,
        draggable: true,
        width:200,
        buttons : {
            "Manual" : function(){
                $('#source-image').imgAreaSelect({ enable: false, hide: true });
                create_cropper(null);
            },
            "3:2" : function(){
                $('#source-image').imgAreaSelect({ enable: false, hide: true });
                create_cropper("3:2");
            },
            "4:3" : function(){
                $('#source-image').imgAreaSelect({ enable: false, hide: true });
                create_cropper("4:3");
            },
            "16:9" : function(){
                $('#source-image').imgAreaSelect({ enable: false, hide: true });
                create_cropper("16:9");
            },
            "5:4" : function(){
                $('#source-image').imgAreaSelect({ enable: false, hide: true });
                create_cropper("5:4");
            },
            "1:1" : function(){
                $('#source-image').imgAreaSelect({ enable: false, hide: true });
                create_cropper("1:1");
            }
        }
    });
    
}

function remove_crop(){
    $('#source-image').imgAreaSelect({ enable: false, hide: true });
    $("#aspect-ratio-selector").dialog('destroy');
}
        
on('action_button_clicked').accomplish(function(btn){

if($(btn).attr('id') == "crop-button"){
    add_crop();
}else{
    remove_crop();
}

});

on('before_image_reload').accomplish(function(){
    remove_crop();
});

on('before_image_zoom_change').accomplish(function(params){
    var active_btn = $('input.edit-button.active');
    if(active_btn.attr('id') == 'crop-button'){
        remove_crop();
    }
});
on('after_image_zoom_change').accomplish(function(params){
    var active_btn = $('input.edit-button.active');
    if(active_btn.attr('id') == 'crop-button'){
        add_crop();
    }
});
        """
    
    def action_parameters(self):
        return """(function(){
    var cs = {};
    var zoom = parseFloat(jQuery("#zoom-slider-value").attr('value'));
    cs['crop.x1'] = window.crop_selection.x1/zoom;
    cs['crop.x2'] = window.crop_selection.x2/zoom;
    cs['crop.y1'] = window.crop_selection.y1/zoom;
    cs['crop.y2'] = window.crop_selection.y2/zoom;
    return cs;
})"""
    
    def __call__(self, x1, y1, x2, y2):
        image = self.editor.get_current_image()
        box = (int(float(x1)), int(float(y1)), int(float(x2)), int(float(y2)))
        new_image = image.crop(box=box)
        new_image.load()
        
        self.editor.set_image(new_image, image.format)
        
        
class RotateLeftAction(BaseImageEditorAction):
    
    implements(IImageEditorAction)
    
    options = form.FormFields(INoOptions)
    
    name = u"Rotate Left"
    description = u"Rotate the image to the left."
    skip_apply = True
    icon = u"++resource++imageeditor/icons/stock-rotate-270-16.png"
    def __call__(self):
        original = self.editor.get_current_image()
        image = original.rotate(90)

        self.editor.set_image(image, original.format)
        
class BlurAction(BaseImageEditorAction):
    implements(IImageEditorAction)
    
    options = form.FormFields(IBlurOptions)
    options['amount'].custom_widget = SliderWidget
    name = u"Blur"
    description = u"Blur this image."
    icon = u"++resource++imageeditor/icons/stock-tool-blur-16.png"
    
    def __call__(self, amount):
        image = self.editor.get_current_image()
        fmt = image.format
        for x in range(0, int(amount)):
            image = image.filter(ImageFilter.BLUR)
            
        self.editor.set_image(image, fmt)
    
class SaveImageEditAction(BaseImageEditorAction):
    implements(IImageEditorAction)
    
    options = form.FormFields(INoOptions)
    
    name = "Save"
    description = "Save the edited image."
    skip_apply = True
    icon = u"++resource++imageeditor/icons/apply.png"
    
    def on_setup(self):
        return """
function save_reload(data){
    var btn = $('#save-button');

    if(data.can_save){
        btn[0].disabled = false;
        btn.removeClass('disabled');
    }else{
        btn[0].disabled = true;
        btn.addClass('disabled');
    }
}
on('after_image_reload').accomplish(save_reload);
save_reload(IMAGE_INFORMATION);
        """
    
    def __call__(self):
        portal_repository = getToolByName(self.editor.context, 'portal_repository')
        if portal_repository.isVersionable(self.editor.context):
            portal_repository.save(self.editor.context, comment = "saved from image editor")

        self.editor.save_edit()
        
        self.editor.context.reindexObject() #stop image caching on browser
    
class SaveAsImageEditAction(BaseImageEditorAction):
    implements(IImageEditorAction)

    options = form.FormFields(ISaveAsOptions)
    
    name = "Save As"
    description = "Save the edited image as another content item."
    icon = None

    def on_setup(self):
        return """
function redirect(data){
    if(data.previous_action == 'save-as'){
        window.location = data.new_type_location;
    }
}
on('after_image_reload').accomplish(redirect);
"""

    def __call__(self, type_to_save_as, title, *args, **kwargs):
        """
        create the new type, pass along the url to the client and then
        the javascript will redirect the browser
        """
        if not title:
            title = self.editor.context.Title()
            
        parent = self.editor.context.getParentNode()
        
        new_id = orig_id = queryUtility(IURLNormalizer).normalize(title)
        
        count = 1
        while new_id in parent.objectIds():
            new_id = orig_id + "-" + str(count)
            count += count
        
        parent.invokeFactory(
            type_to_save_as,
            new_id,
            title = title,
            image = self.editor.get_current_image_data()
        )
        
        return {'new_type_location' : parent.absolute_url() + "/" + new_id + "/edit"}
        
class CancelImageEditAction(BaseImageEditorAction):
    implements(IImageEditorAction)

    options = form.FormFields(INoOptions)

    name = "Cancel"
    description = "Cancel the current edit."
    skip_apply = True
    icon = u"++resource++imageeditor/icons/process-stop.png"
    
    def on_setup(self):
        return """
        
function cancel_reload(data){
    var btn = $('#cancel-button');

    if(data.can_save){
        btn[0].disabled = false;
        btn.removeClass('disabled');
    }else{
        btn[0].disabled = true;
        btn.addClass('disabled');
    }
}
on('after_image_reload').accomplish(cancel_reload);
cancel_reload(IMAGE_INFORMATION);
        """
    
    def __call__(self):
        self.editor.clear_edits()

class RedoAction(BaseImageEditorAction):
    implements(IImageEditorAction)

    options = form.FormFields(INoOptions)
    name = u"Redo"
    description = u"Redo the previous undo action."
    skip_apply = True
    icon = u"++resource++imageeditor/icons/edit-redo.png"
    
    def on_setup(self):
        return """
        
function redo_reload(data){
    var btn = $('#redo-button');

    if(data.can_redo){
        btn[0].disabled = false;
        btn.removeClass('disabled');
    }else{
        btn[0].disabled = true;
        btn.addClass('disabled');
    }
}
        
on('after_image_reload').accomplish(redo_reload);
redo_reload(IMAGE_INFORMATION)
        """
    
    def __call__(self):
        self.editor.redo()
        
class UndoAction(BaseImageEditorAction):
    implements(IImageEditorAction)

    options = form.FormFields(INoOptions)
    name = u"Undo"
    description = u"Go back to the previous change."
    skip_apply = True
    icon = u"++resource++imageeditor/icons/edit-undo.png"

    def on_setup(self):
        return """
function undo_reload(data){
    var btn = $('#undo-button');

    if(data.can_undo){
        btn[0].disabled = false;
        btn.removeClass('disabled');
    }else{
        btn[0].disabled = true;
        btn.addClass('disabled');
    }
}
        
on('after_image_reload').accomplish(undo_reload);
undo_reload(IMAGE_INFORMATION);
        """

    def __call__(self):
        self.editor.undo()
        
class RotateRightAction(BaseImageEditorAction):
    implements(IImageEditorAction)

    options = form.FormFields(INoOptions)
    name = u"Rotate Right"
    description = u"Rotate the image right."
    skip_apply = True
    icon = u"++resource++imageeditor/icons/stock-rotate-90-16.png"
    
    def __call__(self):
        original = self.editor.get_current_image()
        image = original.rotate(270)

        self.editor.set_image(image, original.format)

class FlipOnVerticalAxisAction(BaseImageEditorAction):
    implements(IImageEditorAction)

    options = form.FormFields(INoOptions)
    name = u"Flip Vertical"
    description = u"Flip the image on vertically."
    skip_apply = True
    icon = u"++resource++imageeditor/icons/stock-tool-rotate-16.png"
    
    def __call__(self):
        original = self.editor.get_current_image()
        image = original.transpose(Image.FLIP_TOP_BOTTOM)

        self.editor.set_image(image, original.format)

class CompressAction(BaseImageEditorAction):
    implements(IImageEditorAction)

    options = form.FormFields(ICompressOptions)
    options['amount'].custom_widget = SliderWidget
    name = u"Compress"
    description = u"Compress this image to make it lower quality."
    icon = u"++resource++imageeditor/icons/stock-template-16.png"

    def __call__(self, amount):
        image = self.editor.get_current_image().convert('RGB') # if it is a png, convert it...
        self.editor.set_image(image, quality=float(amount))

class ContrastAction(BaseImageEditorAction):
    implements(IImageEditorAction)
    
    options = form.FormFields(IContrastOptions)
    options['amount'].custom_widget = SliderWidget
    name = u"Contrast"
    description = u"Change the contrast of the image."
    icon = u"++resource++imageeditor/icons/stock-tool-contrast-16.png"
    
    def __call__(self, amount):
        image = self.editor.get_current_image()
        enhancer = ImageEnhance.Contrast(image)
        newImage = enhancer.enhance((float(amount)/100)*2.0)

        self.editor.set_image(newImage, image.format)
    
class BrightnessAction(BaseImageEditorAction):
    implements(IImageEditorAction)
    
    options = form.FormFields(IBrightnessOptions)
    options['amount'].custom_widget = SliderWidget
    name = u"Brightness"
    description = u"Change the brightness of the image."
    icon = u"++resource++imageeditor/icons/stock-tool-brightness-16.png"
    
    def __call__(self, amount):
        image = self.editor.get_current_image()
        enhancer = ImageEnhance.Brightness(image)
        #can enhance from 0.0-2.0, 1.0 being original image
        newImage = enhancer.enhance((float(amount)/100)*2.0)

        self.editor.set_image(newImage, image.format)
        
        
class SharpenAction(BaseImageEditorAction):
    implements(IImageEditorAction)
    
    options = form.FormFields(ISharpenOptions)
    options['amount'].custom_widget = SliderWidget
    name = u"Sharpen"
    description = u"Sharpen the image."
    icon = u"++resource++imageeditor/icons/stock-tool-colorize-16.png"
    
    def __call__(self, amount):
        image = self.editor.get_current_image()
        enhancer = ImageEnhance.Sharpness(image)
        newImage = enhancer.enhance(int(amount))

        self.editor.set_image(newImage, image.format)

class FlipOnHorizontalAxisAction(BaseImageEditorAction):
    implements(IImageEditorAction)
    
    options = form.FormFields(INoOptions)
    name = u"Flip Horizontally"
    description = u"Flip the image on the horizontal axis."
    skip_apply = True
    icon = u"++resource++imageeditor/icons/stock-tool-flip-16.png"
    
    def __call__(self):
        original = self.editor.get_current_image()
        image = original.transpose(Image.FLIP_LEFT_RIGHT)

        self.editor.set_image(image, original.format)

class ResizeAction(BaseImageEditorAction):
    implements(IImageEditorAction)
    
    options = form.FormFields(INoOptions)
    name = u"Resize"
    description = u"Resize the image.."
    icon = u"++resource++imageeditor/icons/stock-resize-16.png"
    
    def on_setup(self):
        return """
function add_resize(){
    $('#image-container')[0].scrollTop = 5000;
    $('#image-container')[0].scrollLeft = 5000;
    
    function create_resizable(){
        $('#source-image').resizable({
            handles: 'all',
            resize: function(e, ui){
                //nothing
            }
        });
    }
    create_resizable();
    
    $('#resize-to-fields').dialog({
        autoOpen:true,
        resizable:false,
        modal: false,
        draggable: true,
        title: "Manual Resize",
        zindex: 1002,
        position: 'right',
        buttons: {
            Resize: function() {
                $('#source-image').resizable('destroy');
                var width = $('#resize-width').val();
                var height = $('#resize-height').val();
                if(isNaN(width) || isNaN(height)){
                    alert("You must enter a number for the width and height.");
                }else{
                    var zoom = parseFloat(jQuery("#zoom-slider-value").attr('value'));
                
                    $('#source-image').width(Math.floor(parseInt(width)*zoom) + "px");
                    $('#source-image').height(Math.floor(parseInt(height)*zoom) + "px");
                }
                
                create_resizable();
            },
            Close: function() {
                $(this).dialog('close');
            }
        }
        
    });
}
function remove_resize(){
    $('#source-image').resizable('destroy');
    $('#resize-to-fields').dialog('destroy');
}
on('action_button_clicked').accomplish(function(btn){

if($(btn).attr('id') == "resize-button"){
    add_resize();
}else{
    remove_resize();
}

});

on('before_image_reload').accomplish(function(){
    remove_resize();
});

on('before_image_zoom_change').accomplish(function(params){
    var active_btn = $('input.edit-button.active');
    if(active_btn.attr('id') == 'resize-button'){
        remove_resize();
    }
});
on('after_image_zoom_change').accomplish(function(params){
    var active_btn = $('input.edit-button.active');
    if(active_btn.attr('id') == 'resize-button'){
        add_resize();
    }
});
            """

    def action_parameters(self):
        return """(function(){
    var res = {};
    var zoom = parseFloat(jQuery("#zoom-slider-value").attr('value'));
    
    res['resize.width'] = jQuery('#source-image').width()/zoom;
    res['resize.height'] = jQuery('#source-image').height()/zoom;
    return res;
})"""
    
    def __call__(self, width, height):
        image = self.editor.get_current_image()
        size=(int(float(width)), int(float(height)))
        new_image = image.resize(size, Image.ANTIALIAS)
        
        self.editor.set_image(new_image, image.format)

class DropShadowAction(BaseImageEditorAction):
    implements(IImageEditorAction)
    
    options = form.FormFields(IDropShadowOptions)
    name = u"Drop Shadow"
    description = u"Adds a drop shadow to the image."
    icon = u"++resource++imageeditor/icons/stock-transparency-16.png"
    
    def __call__(self, offset_x, offset_y, background_color, shadow_color, border, iterations):
        image = self.editor.get_current_image().convert('RGB') #convert to png if it isn't--shadow won't work without this.
        offset = (int(offset_x), int(offset_y))
        background=eval("0x" + background_color[:6]) #precaution in case code is attempted to be injected
        shadow=eval("0x" + shadow_color[:6]) #precaution in case code is attempted to be injected
        border=int(border)
        iterations=int(iterations)
        
        # Create the backdrop image -- a box in the background colour with a 
        # shadow on it.
        totalWidth = image.size[0] + abs(offset[0]) + 2*border
        totalHeight = image.size[1] + abs(offset[1]) + 2*border
        back = Image.new(image.mode, (totalWidth, totalHeight), background)

        # Place the shadow, taking into account the offset from the image
        shadowLeft = border + max(offset[0], 0)
        shadowTop = border + max(offset[1], 0)
        back.paste(
            shadow, 
            [shadowLeft, shadowTop, shadowLeft + image.size[0], shadowTop + image.size[1]] 
        )

        # Apply the filter to blur the edges of the shadow.  Since a small kernel
        # is used, the filter must be applied repeatedly to get a decent blur.
        for n in range(0, iterations):
            back = back.filter(ImageFilter.BLUR)

        # Paste the input image onto the shadow backdrop  
        imageLeft = border - min(offset[0], 0)
        imageTop = border - min(offset[1], 0)
        back.paste(image, (imageLeft, imageTop))
        
        self.editor.set_image(back)

class SepiaAction(BaseImageEditorAction):
    implements(IImageEditorAction)
    
    options = form.FormFields(ISepiaOptions)
    options['red'].custom_widget = SliderWidget
    options['green'].custom_widget = SliderWidget
    options['blue'].custom_widget = SliderWidget
    
    name = u"Sepia"
    description = u"Applies the sepia effect to the image."
    icon = None
    
    def make_linear_ramp(self, white):
        # putpalette expects [r,g,b,r,g,b,...]
        ramp = []
        r, g, b = white
        for i in range(255):
            ramp.extend((r*i/255, g*i/255, b*i/255))
        return ramp        
    
    def __call__(self, red, green, blue):
        """
        found at http://effbot.org/zone/pil-sepia.htm
        """
        sepia = self.make_linear_ramp((int(red), int(green), int(blue)))
        image = self.editor.get_current_image()

        # convert to grayscale
        if image.mode != "L":
            image = image.convert("L")

        # optional: apply contrast enhancement here, e.g.
        image = ImageOps.autocontrast(image)

        # apply sepia palette
        image.putpalette(sepia)

        # convert back to RGB so we can save it as JPEG
        # (alternatively, save it in PNG or similar)
        image = image.convert("RGB")
        
        self.editor.set_image(image)


from Products.ATContentTypes.interface.image import IImageContent
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.ImageEditor.interfaces import *
from Products.ImageEditor.meta.zcml import *
from Products.ImageEditor.utils import *
from plone.memoize.view import memoize
from time import gmtime, strftime
from zope.interface import implements

class ShowCurrentEdit(BrowserView):
    """
    Just a browser view to get latest edited image
    """
    def __init__(self, context, request):
        super(BrowserView, self).__init__(context, request)
        self.imageeditor = IImageEditorAdapter(context)

    def __call__(self):
        
        resp = self.request.response
        imagedata = self.imageeditor.get_current_image_data()
        resp.setHeader('Content-Type', 'image/jpeg')
        resp.setHeader('Content-Length', len(imagedata))
        resp.setHeader('Last-Modified', strftime('%a, %d %b %Y %H:%M:%S +0000', gmtime()))
        resp.write(imagedata)
        return ''


class ImageEditorActionExecute(BrowserView):
    
    def get_args_from_request(self, action):
        args = {}
        
        for key in self.request.keys():
            if key.startswith(action + "."):
                args[key.split('.')[1]] = self.request.get(key)
                
        return args
    
    def __call__(self, action_name):
        #get action instance
        action = get_action_class(action_name)(self.context)
        #call instance with params
        result = action(**self.get_args_from_request(action_name)) or {}
        
        result.update(get_image_information(action.editor))
        result['previous_action'] = action_name
        
        return json(result)


class ImageEditorUtility(BrowserView):
    """Traversable utility for image editor
    """
    implements(IImageEditorUtility)

    @memoize
    def editable(self):
        pm = getToolByName(self.context, 'portal_membership')
        
        return IImageContent.providedBy(self.context) and not \
                pm.isAnonymousUser()

    @memoize
    def should_include(self, request):
        pm = getToolByName(self.context, 'portal_membership')

        return IImageContent.providedBy(self.context) and \
                not pm.isAnonymousUser() and \
                '/@@imageeditor.' in request.get('ACTUAL_URL')



from Testing.ZopeTestCase import installPackage
from Products.Five import zcml, fiveconfigure
from collective.testcaselayer.ptc import BasePTCLayer, ptc_layer


class Layer(BasePTCLayer):
    """ basic layer for integration tests """

    def afterSetUp(self):
        # load zcml...
        fiveconfigure.debug_mode = True
        from Products import ImageEditor
        zcml.load_config('testing.zcml', package=ImageEditor)
        fiveconfigure.debug_mode = False
        # initialize packages...
        installPackage('Products.ImageEditor', quiet=True)
        installPackage('collective.js.jquery', quiet=True)
        installPackage('collective.js.jqueryui', quiet=True)
        # quick-install...
        self.addProfile('Products.ImageEditor:default')


basic = Layer(bases=[ptc_layer])



from actions import CropAction, RotateLeftAction, BlurAction, \
    SaveImageEditAction, SaveAsImageEditAction, CancelImageEditAction, \
    RedoAction, UndoAction, RotateRightAction, FlipOnVerticalAxisAction, \
    CompressAction, ContrastAction, BrightnessAction, SharpenAction, \
    FlipOnHorizontalAxisAction, ResizeAction, DropShadowAction, \
    SepiaAction
    
from base import BaseImageEditorAction

from options import INoOptions, IBlurOptions, ICompressOptions, \
    IContrastOptions, IBrightnessOptions, ISharpenOptions, \
    IDropShadowOptions, ISepiaOptions, ISaveAsOptions
    
from widgets import SliderWidget


from plone.memoize.view import memoize
from Products.Five.browser import BrowserView
from Products.ImageEditor.interfaces.imageeditor import IImageEditorAdapter
from Products.ImageEditor.meta.zcml import get_actions
from Products.ImageEditor.utils import generate_random_url, get_image_information, json
from zope.formlib import form
from Products.ImageEditor import imageeditor_message_factory as _
from Products.CMFCore.utils import getToolByName

class ImageEditor(BrowserView):
    """Redirect based on the user preferences"""
    def __call__(self, *args, **kwargs):
        pm = getToolByName(self.context, 'portal_membership')
        member = pm.getAuthenticatedMember()
        pref = 'alagimp'
        if member is not None:
            pref = member.getProperty('image_editor', 'alagimp')
        self.request.response.redirect('@@imageeditor.%s'%pref)
        return ''

class Base(BrowserView):
    """Basic view for image editor"""

    def __init__(self, context, request):
        super(BrowserView, self).__init__(context, request)
        self.editor = IImageEditorAdapter(self.context)
        self.editor.set_field(request.get('field'))
        self.actions = [(name, action.class_(self.context)) for name, action in get_actions()]

    def get_buttons(self):
        buttons = []
        for name, action in self.actions:
            info = {
                'id': name + '-button',
                'value' : action.name,
                'name' : name,
                'alt' : action.description,
            }
            if action.icon:
                info['style'] = "background-image: url(%s)" % action.icon
            buttons.append(info)
        return buttons

    def get_options(self):
        html = ''
        for name, action in self.actions:
            html += '<div class="image-edit-action" id="%s-options">' % name
            widgets = form.setUpInputWidgets(
                action.options, 
                name,
                self.context,
                self.request,
                ignore_request=True
            )
            
            for widget in widgets:
                html += """
<div class="edit-option">
    <label class="formQuestion" for="%s.%s">%s</label>
    <div class="formHelp">%s</div>
    %s
</div>
                """ % (
                    name,
                    widget.name,
                    widget.context.title.default,
                    widget.context.description.default,
                    widget()
                )
            
            if not action.skip_apply:
                html += """
<input type="button" id="%(name)s-apply-button" 
       class="image-edit-apply-button" name="%(name)s" 
       value="Apply" 
/>
                """ % {'name': _(name)}
            
            html += '</div>'
            
        return html

    def setup_js(self):
        setup_js = []
        
        for name, action in self.actions:
            js = action.on_setup()
            if js:
                setup_js.append(js)

        return """
var IMAGE_INFORMATION = %s;
(function($){
$(document).ready(function(){

%s

});
})(jQuery);
        """ % (json(get_image_information(self.editor)), '\n'.join(setup_js))
        
    def custom_action_parameters(self):
        params = []
        
        for name, action in self.actions:
            ap = action.action_parameters()
            if ap:
                params.append("ACTION_PARAMETERS['%s'] = %s;" % (name, ap))
        
        return """
var ACTION_PARAMETERS = {};
%s    
        """ % '\n'.join(params)

    @memoize
    def image_url(self):
        """
        This is used because sometimes browsers cache images that may have been edited
        """
        return generate_random_url(self.context)



from zope.component import Interface
from zope import schema
from Products.ImageEditor import imageeditor_message_factory as _

class INoOptions(Interface):
    """
    when there are no options available....
    """
    
class IBlurOptions(Interface):
    """
    """
    
    amount = schema.Int(
        title=_(u"label_blur_amount_title", default=u"Blur Amount"),
        description=_(u"label_blur_amount_description",
            default=u"Choose the level you want to blur the image"),
        min=0,
        max=8,
        default=2
    )
    
class ICompressOptions(Interface):
    
    amount = schema.Float(
        title=_(u"label_compress_amount_title", default=u"Compress Amount"),
        description=_(u"label_compress_amount_description", 
            default=u"How far to compress"),
        min=0.0,
        max=100.0,
        default=60.0
    )

class IContrastOptions(Interface):

    amount = schema.Float(
        title=_(u"label_contrast_amount_title", default=u"Amount"),
        description=_(u"label_contrast_amount_description", 
            default=u"Number between 0 and 100 to set the contrast to."),
        min=0.0,
        max=100.0,
        default=50.0
    )
    
class IBrightnessOptions(Interface):
    
    amount = schema.Float(
        title=_(u"label_brightness_amount_title", default=u"Amount"),
        description=_(u"label_brightness_amount_description",
            default=u"The Amount the brightness should be adjusted"),
        min=0.0,
        max=100.0,
        default=50.0
    )
    
    
class ISharpenOptions(Interface):
    
    amount = schema.Int(
        title=_(u"label_sharpen_amount_title", default=u"Amount"),
        description=_(u"label_sharpen_amount_description",
            default=u"The amount to sharpen the image"),
        min=1,
        max=10,
        default=2
    )
    
class IDropShadowOptions(Interface):
    
    offset_x = schema.Int(
        title=_(u"label_dropshadow_offset_x_title", default=u"Offset X"),
        description=_(u"label_dropshadow_offset_x_description",
            default=u"The offset the shadow should have on the x axis"),
        default=5
    )
    
    offset_y = schema.Int(
        title=_(u"label_dropshadow_offset_y_title", default=u"Offset Y"),
        description=_(u"label_dropshadow_offset_y_description",
            default=u"The offset the shadow should have on the y axis"),
        default=5
    )
    
    background_color = schema.TextLine(
        title=_(u"label_dropshadow_background_color_title", default=u"Background color"),
        description=_(u"label_dropshadow_background_color_description",
            default=u"The hex of the background color to use"),
        default=u"ffffff"
    )
    
    shadow_color = schema.TextLine(
        title=_(u"label_dropshadow_shadow_color_title", default=u"Shadow Color"),
        description=_(u"label_dropshadow_shadow_color_description",
            default=u"The hex of the color that will be used for the shadow."),
        default=u"444444"
    )
    
    border = schema.Int(
        title=_(u"label_dropshadow_border_title", default=u"Width"),
        description=_(u"label_dropshadow_border_description",
            default=u"How wide the shadow will be."),
        default=8
    )
    
    iterations = schema.Int(
        title=_(u"label_dropshadow_iterations_title", default=u"Softness"),
        description=_(u"label_dropshadow_iterations_description",
            default=u"Affects how heavy the shadow appears."),
        default=3
    )
    
class ISepiaOptions(Interface):
    
    red = schema.Int(
        title=_(u"label_sepia_red", default=u"Red"),
        description=_(u"label_description_sepia_red", 
            default=u"The level for red to apply for the sepia effect."),
        min=0,
        max=255,
        default=255
    )
    
    green = schema.Int(
        title=_(u"label_sepia_green", default=u"Green"),
        description=_(u"label_description_sepia_green", 
            default=u"The level for green to apply for the sepia effect."),
        min=0,
        max=255,
        default=240
    )
    
    blue = schema.Int(
        title=_(u"label_sepia_blue", default=u"Blue"),
        description=_(u"label_description_sepia_blue", 
            default=u"The level for blue to apply for the sepia effect."),
        min=0,
        max=255,
        default=192
    )
    
class ISaveAsOptions(Interface):
    
    type_to_save_as = schema.Choice(
        title=_(u"label_type_to_save_as", default=u"Type"),
        description=_(u"label_description_type_to_save_as",
            default=u"Content type to save image as."),
        default="Image",
        required=True,
        vocabulary="plone.app.vocabularies.ImageContentTypes"
    )
    
    title = schema.TextLine(
        title=_(u"label_title_save_as", default=u"Title"),
        description=_(u"label_description_title_save_as",
            default=u"Title of the content type that is also used to generate its id."),
        required=True
    )


from Products.CMFCore.utils import getToolByName

def uninstall_import(context):
    
    if context.readDataFile('Products.ImageEditor.txt') is None:
        return

    portal = context.getSite()
    catalog = getToolByName(portal, 'portal_catalog')
    
    images = catalog.searchResults(portal_type=["Image"])
    
    for image in images:
        image = image.getObject()
        
        if hasattr(image, 'stack_pos'):
            delattr(image, 'stack_pos')
            
        if hasattr(image, 'unredostack'):
            delattr(image, 'unredostack')
            
        image._p_changed = 1
        
        
    portal_actions = getToolByName(portal, 'portal_actions')
    object_buttons = portal_actions.object

    actions_to_remove = ('image_editor',)
    for action in actions_to_remove:
        if action in object_buttons.objectIds():
            object_buttons.manage_delObjects([action])


from setuptools import setup, find_packages
from os.path import join

version = open(join('Products', 'ImageEditor', 'version.txt')).read()
readme = open("README.txt").read()
history = open(join('docs', 'HISTORY.txt')).read()

setup(name = 'Products.ImageEditor',
      version = version,
      description = 'adds a "Image Editor" link near the image widget '
        ' allowing the user to rotate, flip, '
        'blur, compress, change contrast & brightness, sharpen, add drop '
        'shadows, crop, resize an image, save as, and apply sepia.',
      long_description = readme + '\n' + history,
      classifiers = [
        "Framework :: Plone",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
      ],
      keywords='plone image editor rotate flip blur compress contrast '
        'brightness sharpen drop shadows crop resize save as sepia',
      author = 'Nathan Van Gheem',
      author_email = 'vangheem@gmail.com',
      url = 'http://www.plone.org/products/products-imageeditor',
      license = 'GPL',
      packages = find_packages(exclude=['ez_setup']),
      namespace_packages = ['Products'],
      include_package_data = True,
      zip_safe = False,
      install_requires = [
        'setuptools',
        'collective.javascript.jqueryui',
      ],
      extras_require = { 'test': [
        'collective.testcaselayer',
      ]},
      entry_points = '''
        [z3c.autoinclude.plugin]
        target = plone
      ''',
)



from StringIO import StringIO
from os.path import dirname, join
from PIL import Image
from Products.PloneTestCase import ptc
from Products.CMFCore.utils import getToolByName
from Products.ImageEditor import testing, tests
from Products.ImageEditor.adapters.imageeditor import ImageEditorAdapter


ptc.setupPloneSite()
ptc.utils.setupCoreSessions()


class ImageEditorTestCase(ptc.PloneTestCase):
    """ base class for integration tests """

    layer = testing.basic

    def afterSetUp(self):
        self.setRoles(('Manager',))
        # put SESSION object into REQUEST
        sdm = self.app.session_data_manager
        self.app.REQUEST.set('SESSION', sdm.getSessionData())

    def getImageContentType(self):
        try:
            images_folder = self.portal.invokeFactory(type_name="Folder", id="images_folder")
        except Exception:
            images_folder = "images_folder"

        try:

            id = self.portal['images_folder'].invokeFactory(type_name="Image", id="testimage")
        except Exception:
            id = 'testimage'
        image = self.portal[images_folder][id]
        image.setTitle('test')
        #(Pdb) p context.getField('image').get(context)
        #<Image at /plone/bug-mantis-01.png/image>
        #(Pdb) p context.getField('image').get(context).data
        #<OFS.Image.Pdata object at 0xb277db6c>
        #(Pdb) p context.getField('image').get(context).data.data
        #'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x04N\x00\x00...
        #...
        #
        #In testcase:
        #(Pdb) image.getField('image').get(image)
        #<Image at /plone/testimage/image>
        #(Pdb) image.getField('image').get(image).data
        #'\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\...
        im = self.getOriginal()
        imageData = StringIO()
        im.save(imageData, im.format)
        image.setImage(imageData)
        return image

    def uninstall(self):
        setup_tool = getToolByName(self.portal, 'portal_setup')
        setup_tool.runAllImportStepsFromProfile('profile-Products.ImageEditor:uninstall')

    def getEditor(self, context):
        return ImageEditorAdapter(context)

    def getImageEditorAdapter(self):
        image = self.getImageContentType()
        return ImageEditorAdapter(image)

    def getOriginal(self):
        return self.getImage('original.jpg')

    def getImagePath(self, name):
        return join(dirname(tests.__file__), 'testimages', name)

    def getImage(self, name):
        return Image.open(self.getImagePath(name))



from Products.CMFCore.utils import getToolByName

default_profile = 'profile-Products.ImageEditor:default'
to_1_2_profile = 'profile-Products.ImageEditor:upgrade_to_1_2'

def remove_stack_pos_and_unredostack_from_images_in_catalog(catalog):
    images = catalog.searchResults(portal_type=["Image"])

    for image in images:
        image = image.getObject()
    
        if hasattr(image, 'stack_pos'):
            delattr(image, 'stack_pos')
        
        if hasattr(image, 'unredostack'):
            delattr(image, 'unredostack')
        
        image._p_changed = 1
    
def remove_image_editor_tab(portal):
    portal_actions = getToolByName(portal, 'portal_actions')
    object_buttons = portal_actions.object

    actions_to_remove = ('image_editor',)
    for action in actions_to_remove:
        if action in object_buttons.objectIds():
            object_buttons.manage_delObjects([action])

def remove_visual_editor_tab(portal):
    portal_types = getToolByName(portal, 'portal_types')
    actions = list(portal_types.Image._actions[:])

    if 'object/editor' in [a.category + "/" + a.id for a in actions]:
        action = None
        for a in actions:
            if a.id == 'editor' and a.category == 'object':
                action = a
                break
            
        actions.remove(action)
        portal_types.Image._actions = tuple(actions)
        portal_types.Image._p_changed = 1
        

def upgrade_to_1_2(context):
    portal = getToolByName(context, 'portal_url').getPortalObject()
    
    catalog = getToolByName(portal, 'portal_catalog')
    remove_stack_pos_and_unredostack_from_images_in_catalog(catalog)
    remove_image_editor_tab(portal)
    remove_visual_editor_tab(portal)
    
    context.runAllImportStepsFromProfile(to_1_2_profile)
    context.runAllImportStepsFromProfile(default_profile)
    
def upgrade_to_1_3(context):
    context.runImportStepFromProfile(default_profile, 'jsregistry')
    


from Products.ImageEditor.utils import normalize_name
from zope.interface import Interface
from zope.configuration.fields import GlobalObject
from zope.schema import TextLine
from Products.ImageEditor.interfaces.actions import IImageEditorAction


class IEditorAction(Interface):
    """
    Fields to create an image editor action
    """

    class_ = GlobalObject(
        title=u"Class",
        description=u"The actual class that will be called when this action is selected",
        required=True
    )
    
    before = TextLine(
        title=u"Before",
        description=u"Insert this action before another action.",
        required=False
    )
    
    after = TextLine(
        title=u"After",
        description=u"Insert this action after another action.",
        required=False
    )
    

IMAGE_EDITOR_ACTIONS = {}
ACTION_ORDER = []

def get_actions():
    for name in ACTION_ORDER:
        yield name, IMAGE_EDITOR_ACTIONS[name]
    
def get_action_class(name):
    if IMAGE_EDITOR_ACTIONS.has_key(name):
        return IMAGE_EDITOR_ACTIONS[name].class_
    else:
        raise Exception("Can't find action %s" % name)
        
def get_action_names():
    for name in ACTION_ORDER:
        yield name
        
class ImageEditAction:

    def __init__(self, class_, before=None, after=None):
        self.name = normalize_name(class_.name)
        self.class_ = class_
        self.before = before
        self.after = after


def add_action(_context, class_, before=None, after=None):
    
    action = ImageEditAction(class_, before, after)
    
    IMAGE_EDITOR_ACTIONS[action.name] = action
    
    if not IImageEditorAction.implementedBy(action.class_):
        raise Exception("Class must implement IImageEditorAction")
    
    if action.before is not None:
        if action.before == "*":
            ACTION_ORDER.insert(0, action.name)
        else:
            for index in range(0, len(ACTION_ORDER)):
                if action.before == ACTION_ORDER[index]:
                    ACTION_ORDER.insert(index, action.name)
                    break
                    
    elif action.after is not None:
        if action.after == "*":
            ACTION_ORDER.insert(len(ACTION_ORDER), action.name)
        else:
            for index in range(0, len(ACTION_ORDER)):
                if action.after == ACTION_ORDER[index]:
                    ACTION_ORDER.insert(index+1, action.name)
                    break
                    
    else:
        ACTION_ORDER.append(action.name)
    